"""
summary.py
"""
import os


class Summary:
    """ Summary Class for an Mp4File Class """

    def __init__(self, mp4file):
        boxes = mp4file.children
        self.data = {}
        self.data['filename'] = mp4file.filename
        self.data['filesize (bytes)'] = os.path.getsize(mp4file.filename)
        if [box for box in boxes if box.type == 'ftyp' or box.type == 'styp']:
            fstyp = [box for box in boxes if box.type == 'ftyp' or box.type == 'styp'][0]
            self.data['brand'] = fstyp.box_info['major_brand']
        # check if there is a moov and if there is a moov that contains traks N.B only ever 0,1 moov boxes
        if [box for box in boxes if box.type == 'moov']:
            moov = [box for box in boxes if box.type == 'moov'][0]
            mvhd = [mvbox for mvbox in moov.children if mvbox.type == 'mvhd'][0]
            self.data['creation_time'] = mvhd.box_info['creation_time']
            self.data['modification_time'] = mvhd.box_info['modification_time']
            self.data['duration (secs)'] = round(mvhd.box_info['duration'] / mvhd.box_info['timescale'])
            if self.data['duration (secs)'] > 0:
                self.data['bitrate (bps)'] = round(8 * self.data['filesize (bytes)'] / self.data['duration (secs)'])
            traks = [tbox for tbox in moov.children if tbox.type == 'trak']
            self.data['track_list'] = []
            for trak in traks:
                this_trak = {}
                this_trak['track_id'] = [box for box in trak.children if box.type == 'tkhd'][0].box_info['track_ID']
                mdia = [box for box in trak.children if box.type == 'mdia'][0]
                mdhd = [box for box in mdia.children if box.type == 'mdhd'][0]
                hdlr = [box for box in mdia.children if box.type == 'hdlr'][0]
                stbl = [box for box in [box for box in mdia.children if box.type == 'minf'][0].children
                             if box.type == 'stbl'][0]
                t = mdhd.box_info['timescale']
                d = mdhd.box_info['duration']
                v = mdhd.version

                sz = [box for box in stbl.children if box.type == 'stsz' or box.type == 'stz2'][0]
                sc = sz.box_info['sample_count']
                if sz.box_info['sample_size'] > 0:
                    # uniform sample size
                    trak_size = sz.box_info['sample_size'] * sc
                else:
                    trak_size = sum([entry['entry_size'] for entry in sz.box_info['entry_list']])

                sample_rate = None
                if (d < 0xffffffff and v == 0) or (d < 0xffffffffffffffff and v == 1):
                    this_trak['track_duration (secs)'] = round(d / t)
                    if trak_size > 0 and this_trak['track_duration (secs)'] > 0:
                        this_trak['track_bitrate (calculated bps)'] = round(8 * trak_size / this_trak['track_duration (secs)'])
                        sample_rate = round((sc * t) / d, 2)

                codec_info = ([box for box in stbl.children if box.type == 'stsd'][0]).children[0]
                media = hdlr.box_info['handler_type']
                if media == 'vide':
                    this_trak['media_type'] = 'video'
                    this_trak['codec_type'] = codec_info.type
                    this_trak['width'] = codec_info.box_info['width'] if 'width' in codec_info.box_info else "unknown"
                    this_trak['height'] = codec_info.box_info['height'] if 'height' in codec_info.box_info else "unknown"
                    if sample_rate is not None:
                        this_trak['frame_rate'] = sample_rate
                elif media == 'soun':
                    this_trak['media_type'] = 'audio'
                    this_trak['codec_type'] = codec_info.type
                    this_trak['channel_count'] = codec_info.box_info['audio_channel_count'] \
                        if 'audio_channel_count' in codec_info.box_info else "unknown"
                    this_trak['sample_rate'] = codec_info.box_info['audio_sample_rate'] \
                        if 'audio_sample_rate' in codec_info.box_info else "unknown"
                else:
                    this_trak['media_type'] = media
                    this_trak['codec_type'] = codec_info.type

                self.data['track_list'].append(this_trak)
        else:
            # no moov found
            self.data['contains_moov'] = False

        if [box for box in boxes if box.type == 'moof']:
            self.data['contains_fragments'] = True






"""
util.py

Utility functions to save me typing struct.unpack all the time.

"""
import struct


def read_u8(fp):
    return struct.unpack('>B', fp.read(1))[0]


def read_u16(fp):
    return struct.unpack('>H', fp.read(2))[0]


def read_u32(fp):
    return struct.unpack('>I', fp.read(4))[0]


def read_u64(fp):
    return struct.unpack('>Q', fp.read(8))[0]


def read_i8(fp):
    return struct.unpack('>b', fp.read(1))[0]


def read_i16(fp):
    return struct.unpack('>h', fp.read(2))[0]


def read_i32(fp):
    return struct.unpack('>i', fp.read(4))[0]


def read_i64(fp):
    return struct.unpack('>q', fp.read(8))[0]


def read_u8_8(fp):
    ipart, fpart = struct.unpack('>2B', fp.read(2))
    return ipart + (fpart / 256)


def read_u16_16(fp):
    ipart, fpart = struct.unpack('>2H', fp.read(4))
    return ipart + (fpart / 65536)


def read_i8_8(fp):
    ipart = struct.unpack('>b', fp.read(1))[0]
    fpart = struct.unpack('B', fp.read(1))[0]
    return ipart + (fpart / 256)





import numpy as np
import pandas as pd
import cv2
import pathlib
import bisect


class CameraParams:
    def __init__(self,
                 # camera info
                 resolution: np.ndarray,
                 # intrinsics
                 camera_mtx: np.ndarray, distort_coeffs: np.ndarray = None,
                 # extrinsics
                 rotation_vec: np.ndarray = None, position: np.ndarray = None):

        self.resolution     : np.ndarray = resolution
        self.camera_mtx     : np.ndarray = camera_mtx
        self.distort_coeffs : np.ndarray = distort_coeffs
        self.rotation_vec   : np.ndarray = rotation_vec
        self.position       : np.ndarray = position

    @staticmethod
    def readFromFile(fileName: str|pathlib.Path) -> 'CameraParams':
        fileName = pathlib.Path(fileName)
        if not fileName.is_file():
            return CameraParams(None,None)

        fs = cv2.FileStorage(str(fileName), cv2.FILE_STORAGE_READ)
        resolution      = fs.getNode("resolution").mat()
        cameraMatrix    = fs.getNode("cameraMatrix").mat()
        distCoeff       = fs.getNode("distCoeff").mat()
        # camera extrinsics for 3D gaze
        cameraRotation  = fs.getNode("rotation").mat()
        if cameraRotation is not None:
            cameraRotation  = cv2.Rodrigues(cameraRotation)[0]  # need rotation vector, not rotation matrix
        cameraPosition  = fs.getNode("position").mat()
        fs.release()

        return CameraParams(resolution,cameraMatrix,distCoeff,cameraRotation,cameraPosition)

    def has_intrinsics(self):
        return (self.camera_mtx is not None) and (self.distort_coeffs is not None)


class CV2VideoReader:
    def __init__(self, file: str|pathlib.Path, timestamps: list|np.ndarray|pd.DataFrame):
        self.file = pathlib.Path(file)
        if isinstance(timestamps,list):
            self.ts = np.array(timestamps)
        elif isinstance(timestamps, pd.DataFrame):
            self.ts = timestamps['timestamp'].to_numpy()
        else:
            self.ts = timestamps

        self.cap = cv2.VideoCapture(str(self.file))
        if not self.cap.isOpened():
            raise RuntimeError('the file "{}" could not be opened'.format(str(self.file)))
        self.nframes= float(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))
        self.frame_idx = -1
        self._last_good_ts = (-1, -1., -1.)  # frame_idx, ts from opencv, ts from file
        self._is_off_by_one = False

    def __del__(self):
        self.cap.release()

    def get_prop(self, cv2_prop):
        return self.cap.get(cv2_prop)

    def set_prop(self, cv2_prop, val):
        return self.cap.set(cv2_prop, val)

    def read_frame(self, report_gap=False):
        ts0 = self.cap.get(cv2.CAP_PROP_POS_MSEC)
        ret, frame = self.cap.read()
        ts1 = self.cap.get(cv2.CAP_PROP_POS_MSEC)
        self.frame_idx += 1

        # check if this is a stream for which opencv returns timestamps that are one frame off
        if self.frame_idx==0 and ts0==ts1:
            self._is_off_by_one = True

        # check if we're done. Can't trust ret==False to indicate we're at end of video, as
        # it may also return False for some corrupted frames that we can just read past
        if not ret and (self.frame_idx==0 or self.frame_idx/self.nframes>.99):
            return True, None, None, None

        # keep going
        ts_from_list = self.ts[self.frame_idx]
        if self.frame_idx==1 or ts1>0.:
            # check for gap, and if there is a gap, fix up frame_idx if needed
            if self._last_good_ts[0]!=-1 and ts_from_list-self._last_good_ts[2] < ts1-self._last_good_ts[1]-1:  # little bit of leeway for precision or mismatched timestamps
                # we skipped some frames altogether, need to correct current frame_idx
                t_jump = ts1-self._last_good_ts[1]
                tss = self.ts-self._last_good_ts[2]
                # find best matching frame idx so we catch up with the jump
                idx = bisect.bisect(tss, t_jump)
                if abs(tss[idx-1]-t_jump)<abs(tss[idx]-t_jump):
                    idx -= 1
                self.frame_idx = idx
                ts_from_list = self.ts[self.frame_idx]
                if report_gap and self.frame_idx-self._last_good_ts[0]>1:
                    print(f'Frame discontinuity detected (jumped from {self._last_good_ts[0]} to {self.frame_idx}), there are probably corrupt frames in your video')
            self._last_good_ts = (self.frame_idx, ts1, ts_from_list)

        # we might not have a valid frame, but we're not done yet
        if not ret or frame is None:
            return False, None,  self.frame_idx, ts_from_list
        else:
            return False, frame, self.frame_idx, ts_from_list

    def report_frame(self, interval=100):
        if self.frame_idx%interval==0:
            print('  frame {}'.format(self.frame_idx))


import numpy as np
import datetime
import csv
import warnings
import bisect

from . import utils

class Timestamp:
    def __init__(self, unix_time: int | float, format="%Y-%m-%d %H:%M:%S"):
        self.format = format
        self.display = ""
        self.value = 0
        self.update(unix_time)

    def update(self, unix_time: int | float):
        self.value = int(unix_time)
        if self.value == 0:
            self.display = ""
        else:
            self.display = datetime.datetime.fromtimestamp(unix_time).strftime(self.format)

utils.register_type(utils.CustomTypeEntry(Timestamp,'__Timestamp__',lambda x: x.value, lambda x: Timestamp(x)))


# for reading frame timestamp files
def from_file(file) -> np.ndarray:
    return np.genfromtxt(file, dtype=None, delimiter='\t', skip_header=1, usecols=1)

class Idx2Timestamp:
    def __init__(self, fileName):
        self.timestamps = {}
        with open(str(fileName), 'r' ) as f:
            reader = csv.DictReader(f, delimiter='\t')
            for entry in reader:
                frame_idx = int(float(entry['frame_idx']))
                if frame_idx!=-1:
                    self.timestamps[frame_idx] = float(entry['timestamp'])

    def get(self, idx):
        if idx in self.timestamps:
            return self.timestamps[int(idx)]
        else:
            warnings.warn('frame_idx %d is not in set\n' % ( idx ), RuntimeWarning )
            return -1.

class Timestamp2Index:
    def __init__(self, fileName):
        self.indices = []
        self.timestamps = []
        with open(str(fileName), 'r' ) as f:
            reader = csv.DictReader(f, delimiter='\t')
            for entry in reader:
                frame_idx = int(float(entry['frame_idx']))
                if frame_idx!=-1:
                    self.indices   .append(int(float(entry['frame_idx'])))
                    self.timestamps.append(    float(entry['timestamp']))

    def find(self, ts):
        idx = bisect.bisect(self.timestamps, ts)
        # return nearest
        if idx>=len(self.timestamps):
            return self.indices[-1]
        elif idx>0 and abs(self.timestamps[idx-1]-ts)<abs(self.timestamps[idx]-ts):
            return self.indices[idx-1]
        else:
            return self.indices[idx]

    def getLast(self):
        return self.indices[-1], self.timestamps[-1]

    def getIFI(self):
        return np.mean(np.diff(self.timestamps))


import numpy as np
import math
import cv2
import pathlib

from . import data_files, drawing, gaze_headref, ocv, plane

class Gaze:
    # description of tsv file used for storage
    _columns_compressed = {'timestamp': 1, 'frame_idx':1,
                           'gazePosCam_vidPos_ray':3,'gazePosCam_vidPos_homography':3,'gazePosCamWorld':3,'gazeOriCamLeft':3,'gazePosCamLeft':3,'gazeOriCamRight':3,'gazePosCamRight':3,
                           'gazePosPlane2D_vidPos_ray':2,'gazePosPlane2D_vidPos_homography':2,'gazePosPlane2DWorld':2,'gazePosPlane2DLeft':2,'gazePosPlane2DRight':2}
    _non_float          = {'frame_idx': int}

    def __init__(self,
                 timestamp                          : float,
                 frame_idx                          : int,

                 gazePosCam_vidPos_ray              : np.ndarray = None,
                 gazePosCam_vidPos_homography       : np.ndarray = None,
                 gazePosCamWorld                    : np.ndarray = None,
                 gazeOriCamLeft                     : np.ndarray = None,
                 gazePosCamLeft                     : np.ndarray = None,
                 gazeOriCamRight                    : np.ndarray = None,
                 gazePosCamRight                    : np.ndarray = None,

                 gazePosPlane2D_vidPos_ray          : np.ndarray = None,
                 gazePosPlane2D_vidPos_homography   : np.ndarray = None,
                 gazePosPlane2DWorld                : np.ndarray = None,
                 gazePosPlane2DLeft                 : np.ndarray = None,
                 gazePosPlane2DRight                : np.ndarray = None):
        # 3D gaze is in world space, w.r.t. scene camera
        # 2D gaze is on the poster
        self.timestamp                          = timestamp
        self.frame_idx                          = frame_idx

        # in camera space (3D coordinates)
        self.gazePosCam_vidPos_ray              = gazePosCam_vidPos_ray             # video gaze position on plane (camera ray intersected with plane)
        self.gazePosCam_vidPos_homography       = gazePosCam_vidPos_homography      # gaze2DHomography in camera space
        self.gazePosCamWorld                    = gazePosCamWorld                   # 3D gaze point on plane (world-space gaze point, turned into direction ray and intersected with plane)
        self.gazeOriCamLeft                     = gazeOriCamLeft
        self.gazePosCamLeft                     = gazePosCamLeft                    # 3D gaze point on plane ( left eye gaze vector intersected with plane)
        self.gazeOriCamRight                    = gazeOriCamRight
        self.gazePosCamRight                    = gazePosCamRight                   # 3D gaze point on plane (right eye gaze vector intersected with plane)

        # in poster space (2D coordinates)
        self.gazePosPlane2D_vidPos_ray          = gazePosPlane2D_vidPos_ray         # Video gaze point mapped to poster by turning into direction ray and intersecting with poster
        self.gazePosPlane2D_vidPos_homography   = gazePosPlane2D_vidPos_homography  # Video gaze point directly mapped to poster through homography transformation
        self.gazePosPlane2DWorld                = gazePosPlane2DWorld               # wGaze3D in poster space
        self.gazePosPlane2DLeft                 = gazePosPlane2DLeft                # lGaze3D in poster space
        self.gazePosPlane2DRight                = gazePosPlane2DRight               # rGaze3D in poster space

    @staticmethod
    def readFromFile(fileName:str|pathlib.Path, start:int=None, end:int=None) -> dict[int,list['Gaze']]:
        return data_files.read_file(fileName,
                                    Gaze, False, False, True,
                                    start=start,end=end)[0]

    @staticmethod
    def writeToFile(gazes: list['Gaze'] | dict[int,list['Gaze']], fileName:str|pathlib.Path, skip_missing=False):
        data_files.write_array_to_file(gazes, fileName,
                                       Gaze._columns_compressed,
                                       skip_all_nan=skip_missing)

    def drawOnWorldVideo(self, img, cameraParams: ocv.CameraParams, subPixelFac=1):
        # project to camera, display
        def project_and_draw(img,pos,sz,clr):
            pPointCam = cv2.projectPoints(pos.reshape(1,3),np.zeros((1,3)),np.zeros((1,3)),cameraParams.camera_mtx, cameraParams.distort_coeffs)[0][0][0]
            if not math.isnan(pPointCam[0]):
                drawing.openCVCircle(img, pPointCam, sz, clr, -1, subPixelFac)

        # gaze ray
        if self.gazePosCam_vidPos_ray is not None:
            project_and_draw(img, self.gazePosCam_vidPos_ray, 3, (255,255,0))
        # binocular gaze point
        if self.gazePosCamWorld is not None:
            project_and_draw(img, self.gazePosCamWorld, 3, (255,0,255))
        # left eye
        if self.gazePosCamLeft is not None:
            project_and_draw(img, self.gazePosCamLeft, 3, (0,0,255))
        # right eye
        if self.gazePosCamRight is not None:
            project_and_draw(img, self.gazePosCamRight, 3, (255,0,0))
        # average
        if (self.gazePosCamLeft is not None) and (self.gazePosCamRight is not None):
            pointCam  = np.array([(x+y)/2 for x,y in zip(self.gazePosCamLeft,self.gazePosCamRight)])
            project_and_draw(img, pointCam, 6, (255,0,255))

    def drawOnPlane(self, img, reference, subPixelFac=1):
        # binocular gaze point
        if self.gazePosPlane2DWorld is not None:
            reference.draw(img, *self.gazePosPlane2DWorld, subPixelFac, (0,255,255), 3)
        # left eye
        if self.gazePosPlane2DLeft is not None:
            reference.draw(img, *self.gazePosPlane2DLeft, subPixelFac, (0,0,255), 3)
        # right eye
        if self.gazePosPlane2DRight is not None:
            reference.draw(img, *self.gazePosPlane2DRight, subPixelFac, (255,0,0), 3)
        # average
        if (self.gazePosPlane2DLeft is not None) and (self.gazePosPlane2DRight is not None):
            average = np.array([(x1+x2)/2 for x1,x2 in zip(self.gazePosPlane2DLeft,self.gazePosPlane2DRight)])
            if not math.isnan(average[0]):
                reference.draw(img, *average, subPixelFac, (255,0,255))
        # video gaze position
        if self.gazePosPlane2D_vidPos_homography is not None:
            reference.draw(img, *self.gazePosPlane2D_vidPos_homography, subPixelFac, (0,255,0), 5)
        if self.gazePosPlane2D_vidPos_ray is not None:
            reference.draw(img, *self.gazePosPlane2D_vidPos_ray, subPixelFac, (255,255,0), 3)


def gazes_head_to_world(poses: list[plane.Pose], gazes_head: dict[int,list[gaze_headref.Gaze]], cameraParams: ocv.CameraParams) -> dict[int,list[Gaze]]:
    from . import transforms
    planeGazes = {}
    for frame_idx in poses:
        if frame_idx in gazes_head:
            planeGazes[frame_idx] = []
            for gaze in gazes_head[frame_idx]:
                gazePoster = transforms.gazeToPlane(gaze, poses[frame_idx], cameraParams)
                planeGazes[frame_idx].append(gazePoster)
    return planeGazes


"""
Cast raw Tobii data into common format.

The output directory will contain:
    - frameTimestamps.tsv: frame number and corresponding timestamps for each frame in video
    - worldCamera.mp4: the video from the point-of-view scene camera on the glasses
    - gazeData.tsv: gaze data, where all 2D gaze coordinates are represented w/r/t the world camera,
                    and 3D coordinates in the coordinate frame of the glasses (may need rotation/translation
                    to represent in world camera's coordinate frame)
    - calibration.xml: info about the camera intrinsics and transform glasses coordinate system to
                       camera coordinate system
"""

import shutil
import pathlib
import json
import gzip
import cv2
import pandas as pd
import numpy as np
import struct
import math
import datetime

from ..recording import Recording
from ..eyetracker import EyeTracker
from .. import timestamps, video_utils


def preprocessData(output_dir: str|pathlib.Path=None, source_dir: str|pathlib.Path=None, rec_info: Recording=None, copy_scene_video = True) -> Recording:
    from . import check_folders
    """
    Run all preprocessing steps on tobii data and store in output_dir
    """
    output_dir, source_dir, rec_info = check_folders(output_dir, source_dir, rec_info, EyeTracker.Tobii_Glasses_2)
    print(f'processing: {source_dir.name} -> {output_dir}')


    ### check and copy needed files to the output directory
    print('  Check and copy raw data...')
    ### check tobii recording and get export directory
    if rec_info is not None:
        checkRecording(source_dir, rec_info)
    else:
        rec_info = getRecordingInfo(source_dir)
        if rec_info is None:
            raise RuntimeError(f"The folder {source_dir} is not recognized as a {EyeTracker.Tobii_Glasses_2.value} recording.")

    # make output dir
    if not output_dir.is_dir():
        output_dir.mkdir()


    ### copy the raw data to the output directory
    srcVid, destVid = copyTobiiRecording(source_dir, output_dir, copy_scene_video)
    if destVid:
        rec_info.scene_video_file = destVid.name
    else:
        rec_info.scene_video_file =  srcVid.parent.parent.name + '/' + srcVid.parent.name + '/' + srcVid.name

    #### prep the copied data...
    print('  Getting camera calibration...')
    sceneVideoDimensions = getCameraFromTSLV(output_dir)
    print('  Prepping gaze data...')
    gazeDf, frameTimestamps = formatGazeData(output_dir, sceneVideoDimensions, rec_info)

    # write the gaze data to a csv file
    gazeDf.to_csv(str(output_dir / 'gazeData.tsv'), sep='\t', na_rep='nan', float_format="%.8f")

    # also store frame timestamps
    frameTimestamps.to_csv(str(output_dir / 'frameTimestamps.tsv'), sep='\t')

    # store rec info
    rec_info.store_as_json(output_dir)

    return rec_info


def getRecordingInfo(inputDir: str|pathlib.Path) -> Recording:
    # returns None if not a recording directory
    recInfo = Recording(source_directory=inputDir, eye_tracker=EyeTracker.Tobii_Glasses_2)

    # get participant info
    file = inputDir / 'participant.json'
    if not file.is_file():
        return None
    with open(file, 'r') as j:
        iInfo = json.load(j)
    recInfo.participant = iInfo['pa_info']['Name']

    # get recording info
    file = inputDir / 'recording.json'
    if not file.is_file():
        return None
    with open(file, 'r') as j:
        iInfo = json.load(j)
    recInfo.name = iInfo['rec_info']['Name']
    recInfo.duration   = int(iInfo['rec_length']*1000)          # in seconds, convert to ms
    time_string = iInfo['rec_created']
    if time_string[-4:].isdigit() and time_string[-5:-4]=='+':
        # add hour:minute separator for ISO 8601 format that datetime understands
        time_string = time_string[:-2]+':'+time_string[-2:]
    recInfo.start_time = timestamps.Timestamp(int(datetime.datetime.fromisoformat(time_string).timestamp()))

    # get system info
    file = inputDir / 'sysinfo.json'
    if not file.is_file():
        return None
    with open(file, 'r') as j:
        iInfo = json.load(j)

    recInfo.firmware_version = iInfo['servicemanager_version']
    recInfo.glasses_serial = iInfo['hu_serial']
    recInfo.recording_unit_serial = iInfo['ru_serial']

    # we got a valid recording and at least some info if we got here
    # return what we've got
    return recInfo


def checkRecording(inputDir: str|pathlib.Path, recInfo: Recording):
    actualRecInfo = getRecordingInfo(inputDir)
    if actualRecInfo is None or recInfo.name!=actualRecInfo.name:
        raise ValueError(f"A recording with the name \"{recInfo.name}\" was not found in the folder {inputDir}.")

    # make sure caller did not mess with recInfo
    if recInfo.participant!=actualRecInfo.participant:
        raise ValueError(f"A recording with the participant \"{recInfo.participant}\" was not found in the folder {inputDir}.")
    if recInfo.duration!=actualRecInfo.duration:
        raise ValueError(f"A recording with the duration \"{recInfo.duration}\" was not found in the folder {inputDir}.")
    if recInfo.start_time.value!=actualRecInfo.start_time.value:
        raise ValueError(f"A recording with the start_time \"{recInfo.start_time.display}\" was not found in the folder {inputDir}.")
    if recInfo.firmware_version!=actualRecInfo.firmware_version:
        raise ValueError(f"A recording with the firmware_version \"{recInfo.firmware_version}\" was not found in the folder {inputDir}.")
    if recInfo.glasses_serial!=actualRecInfo.glasses_serial:
        raise ValueError(f"A recording with the glasses_serial \"{recInfo.glasses_serial}\" was not found in the folder {inputDir}.")
    if recInfo.recording_unit_serial!=actualRecInfo.recording_unit_serial:
        raise ValueError(f"A recording with the recording_unit_serial \"{recInfo.recording_unit_serial}\" was not found in the folder {inputDir}.")


def copyTobiiRecording(inputDir: pathlib.Path, outputDir: pathlib.Path, copy_scene_video: bool):
    """
    Copy the relevant files from the specified input dir to the specified output dir
    """
    # Copy relevent files to new directory
    inputDir = inputDir / 'segments' / '1'
    srcFile  = inputDir / 'fullstream.mp4'
    if copy_scene_video:
        destFile = outputDir / 'worldCamera.mp4'
        shutil.copy2(str(srcFile), str(destFile))
    else:
        destFile = None

    # Unzip the gaze data and tslv files
    for f in ['livedata.json.gz', 'et.tslv.gz']:
        with gzip.open(str(inputDir / f)) as zipFile:
            with open(outputDir / pathlib.Path(f).stem, 'wb') as unzippedFile:
                shutil.copyfileobj(zipFile, unzippedFile)

    return srcFile, destFile

def getCameraFromTSLV(inputDir: str|pathlib.Path):
    """
    Read binary TSLV file until camera calibration information is retrieved
    """
    with open(str(inputDir / 'et.tslv'), "rb") as f:
        # first look for camera item (TSLV type==300)
        while True:
            tslvType= struct.unpack('h',f.read(2))[0]
            status  = struct.unpack('h',f.read(2))[0]
            payloadLength = struct.unpack('i',f.read(4))[0]
            payloadLengthPadded = math.ceil(payloadLength/4)*4
            if tslvType != 300:
                # skip payload
                f.read(payloadLengthPadded)
            else:
                break

        # read info about camera
        camera = {}
        camera['id']       = struct.unpack('b',f.read(1))[0]
        camera['location'] = struct.unpack('b',f.read(1))[0]
        f.read(2) # skip padding
        camera['position'] = np.array(struct.unpack('3f',f.read(4*3)))
        camera['rotation'] = np.reshape(struct.unpack('9f',f.read(4*9)),(3,3))
        camera['focalLength'] = np.array(struct.unpack('2f',f.read(4*2)))
        camera['skew'] = struct.unpack('f',f.read(4))[0]
        camera['principalPoint'] = np.array(struct.unpack('2f',f.read(4*2)))
        camera['radialDistortion'] = np.array(struct.unpack('3f',f.read(4*3)))
        camera['tangentialDistortion'] = np.array(struct.unpack('3f',f.read(4*3))[:-1]) # drop last element (always zero), since there are only two tangential distortion parameters
        camera['resolution'] = np.array(struct.unpack('2h',f.read(2*2)))

    # turn into camera matrix and distortion coefficients as used by OpenCV
    camera['cameraMatrix'] = np.identity(3)
    camera['cameraMatrix'][0,0] = camera['focalLength'][0]
    camera['cameraMatrix'][0,1] = camera['skew']
    camera['cameraMatrix'][1,1] = camera['focalLength'][1]
    camera['cameraMatrix'][0,2] = camera['principalPoint'][0]
    camera['cameraMatrix'][1,2] = camera['principalPoint'][1]

    camera['distCoeff'] = np.zeros(5)
    camera['distCoeff'][:2]  = camera['radialDistortion'][:2]
    camera['distCoeff'][2:4] = camera['tangentialDistortion'][:2]
    camera['distCoeff'][4]   = camera['radialDistortion'][2]


    # store to file
    fs = cv2.FileStorage(str(inputDir / 'calibration.xml'), cv2.FILE_STORAGE_WRITE)
    for key,value in camera.items():
        fs.write(name=key,val=value)
    fs.release()

    # tslv no longer needed, remove
    (inputDir / 'et.tslv').unlink(missing_ok=True)

    return camera['resolution']


def formatGazeData(inputDir: str|pathlib.Path, sceneVideoDimensions: list[int], recInfo: Recording):
    """
    load livedata.json
    format to get the gaze coordinates w/r/t world camera, and timestamps for every frame of video

    Returns:
        - formatted dataframe with cols for timestamp, frame_idx, and gaze data
        - np array of frame timestamps
    """

    # convert the json file to pandas dataframe
    df,scene_video_ts_offset = json2df(inputDir / 'livedata.json', sceneVideoDimensions)

    # read video file, create array of frame timestamps
    frameTimestamps = video_utils.getFrameTimestampsFromVideo(recInfo.get_scene_video_path())
    frameTimestamps['timestamp'] += scene_video_ts_offset

    # use the frame timestamps to assign a frame number to each data point
    frameIdx = video_utils.tssToFrameNumber(df.index,frameTimestamps['timestamp'].to_numpy())
    df.insert(0,'frame_idx',frameIdx['frame_idx'])

    # build the formatted dataframe
    df.index.name = 'timestamp'

    # return the gaze data df and frame time stamps array
    return df, frameTimestamps


def json2df(jsonFile: str|pathlib.Path, sceneVideoDimensions: list[int]) -> tuple[pd.DataFrame, float]:
    """
    convert the livedata.json file to a pandas dataframe
    """
    # dicts to store sync points
    vtsSync  = list()       # scene video timestamp sync
    evtsSync = list()       # eye video timestamp sync (only if eye video was recorded)
    df = pd.DataFrame()     # empty dataframe to write data to

    with open(jsonFile, 'r') as file:
        entries = json.loads('[' + file.read().replace('\n', ',')[:-1] + ']')

    # loop over all lines in json file, each line represents unique json object
    for entry in entries:
        # if non-zero status (error), ensure data found in packet is marked as missing
        isError = False
        if entry['s'] != 0:
            isError = True

        ### a number of different dictKeys are possible, respond accordingly
        if 'vts' in entry.keys(): # "vts" key signfies a scene video timestamp (first frame, first keyframe, and ~1/min afterwards)
            vtsSync.append((entry['ts'], entry['vts'] if not isError else math.nan))
            continue

        ### a number of different dictKeys are possible, respond accordingly
        if 'evts' in entry.keys(): # "evts" key signfies an eye video timestamp (first frame, first keyframe, and ~1/min afterwards)
            evtsSync.append((entry['ts'], entry['evts'] if not isError else math.nan))
            continue

        # if this json object contains "eye" data (e.g. pupil info)
        if 'eye' in entry.keys():
            which_eye = entry['eye'][:1]
            if 'pc' in entry.keys():
                # origin of gaze vector is the pupil center
                df.loc[entry['ts'], 'gaze_ori_'+which_eye+'_x'] = entry['pc'][0] if not isError else math.nan
                df.loc[entry['ts'], 'gaze_ori_'+which_eye+'_y'] = entry['pc'][1] if not isError else math.nan
                df.loc[entry['ts'], 'gaze_ori_'+which_eye+'_z'] = entry['pc'][2] if not isError else math.nan
            elif 'pd' in entry.keys():
                df.loc[entry['ts'], 'pup_diam_'+which_eye] = entry['pd'] if not isError else math.nan
            elif 'gd' in entry.keys():
                df.loc[entry['ts'], 'gaze_dir_'+which_eye+'_x'] = entry['gd'][0] if not isError else math.nan
                df.loc[entry['ts'], 'gaze_dir_'+which_eye+'_y'] = entry['gd'][1] if not isError else math.nan
                df.loc[entry['ts'], 'gaze_dir_'+which_eye+'_z'] = entry['gd'][2] if not isError else math.nan

        # otherwise it contains gaze position data
        else:
            if 'gp' in entry.keys():
                df.loc[entry['ts'], 'gaze_pos_vid_x'] = entry['gp'][0]*sceneVideoDimensions[0] if not isError else math.nan
                df.loc[entry['ts'], 'gaze_pos_vid_y'] = entry['gp'][1]*sceneVideoDimensions[1] if not isError else math.nan
            elif 'gp3' in entry.keys():
                df.loc[entry['ts'], 'gaze_pos_3d_x'] = entry['gp3'][0] if not isError else math.nan
                df.loc[entry['ts'], 'gaze_pos_3d_y'] = entry['gp3'][1] if not isError else math.nan
                df.loc[entry['ts'], 'gaze_pos_3d_z'] = entry['gp3'][2] if not isError else math.nan

        # ignore anything else

    # find out t0. Do the same as GlassesViewer so timestamps are compatible
    # that is t0 is at timestamp of last video start (scene or eye)
    vtsSync  = np.array( vtsSync)
    evtsSync = np.array(evtsSync)
    t0s = [vtsSync[vtsSync[:,1]==0,0]]
    if len(evtsSync)>0:
        t0s.append(evtsSync[evtsSync[:,1]==0,0])
    t0 = max(t0s)

    # get timestamp offset for scene video
    scene_video_ts_offset = (t0s[0]-t0) / 1000.0

    # convert timestamps from us to ms
    df.index = (df.index - t0) / 1000.0

    # json no longer needed, remove
    jsonFile.unlink(missing_ok=True)

    # return the dataframe
    return df, scene_video_ts_offset


"""
Cast raw AdHawk MindLink data into common format.
"""

import shutil
import pathlib
import json
import csv
import cv2
import pandas as pd
import numpy as np
import datetime

from ..recording import Recording
from ..eyetracker import EyeTracker
from .. import timestamps, video_utils


def preprocessData(output_dir: str|pathlib.Path=None, source_dir: str|pathlib.Path=None, rec_info: Recording=None, cam_cal_file: str|pathlib.Path=None, copy_scene_video = True) -> Recording:
    from . import check_folders
    """
    Run all preprocessing steps on AdHawk MindLink data and store in output_dir
    """
    output_dir, source_dir, rec_info = check_folders(output_dir, source_dir, rec_info, EyeTracker.AdHawk_MindLink)
    print(f'processing: {source_dir.name} -> {output_dir}')


    ### check and copy needed files to the output directory
    print('  Check and copy raw data...')
    ### check adhawk recording and get export directory
    if rec_info is not None:
        checkRecording(source_dir, rec_info)
    else:
        rec_info = getRecordingInfo(source_dir)
        if rec_info is None:
            raise RuntimeError(f"The folder {source_dir} is not recognized as a {EyeTracker.AdHawk_MindLink.value} recording.")

    # make output dir
    if not output_dir.is_dir():
        output_dir.mkdir()


    ### copy the raw data to the output directory
    srcVid, destVid = copyAdhawkRecording(source_dir, output_dir, copy_scene_video)
    if destVid:
        rec_info.scene_video_file = destVid.name
    else:
        rec_info.scene_video_file =  srcVid.name

    #### prep the copied data...
    print('  Getting camera calibration...')
    if cam_cal_file is not None:
        shutil.copyfile(str(cam_cal_file), str(output_dir / 'calibration.xml'))
        sceneVideoDimensions = np.array([1280, 720])
    else:
        print('    !! No camera calibration provided! Defaulting to hardcoded')
        sceneVideoDimensions = getCameraHardcoded(output_dir)
    print('  Prepping gaze data...')
    gazeDf, frameTimestamps = formatGazeData(source_dir, sceneVideoDimensions, rec_info)

    # write the gaze data to a csv file
    gazeDf.to_csv(str(output_dir / 'gazeData.tsv'), sep='\t', na_rep='nan', float_format="%.8f")

    # also store frame timestamps
    frameTimestamps.to_csv(str(output_dir / 'frameTimestamps.tsv'), sep='\t')

    # store rec info
    rec_info.store_as_json(output_dir)

    return rec_info


def getRecordingInfo(inputDir: str|pathlib.Path) -> Recording:
    # returns None if not a recording directory
    recInfo = Recording(source_directory=inputDir, eye_tracker=EyeTracker.AdHawk_MindLink)

    # get recording info
    recInfo.name = inputDir.name

    file = inputDir / 'meta_data.json'
    if not file.is_file():
        return None
    with open(file, 'rb') as j:
        rInfo = json.load(j)
    recInfo.duration = rInfo['manifest']['recording_length_ms']
    recInfo.participant = rInfo['user_profile']['name']
    # get recording start time by reading UTC time associated with first gaze sample
    gaze_entry = getMetaEntry(inputDir, 'gaze')
    file = inputDir / gaze_entry['file_name']
    with open(file, 'r') as read_obj:
        csv_reader = csv.DictReader(read_obj)
        # Iterate over each row in the csv using reader object
        sample = next(csv_reader)
    time_string = sample['UTC_Time']
    if time_string[-1:]=='Z':
        # change Z suffix (if any) to +00:00 for ISO 8601 format that datetime understands
        time_string = time_string[:-1]+'+00:00'
    recInfo.start_time = timestamps.Timestamp(int(datetime.datetime.fromisoformat(time_string).timestamp()))

    # we got a valid recording and at least some info if we got here
    # return what we've got
    return recInfo

def getMeta(inputDir: str|pathlib.Path, key: str=None):
    file = inputDir / 'meta_data.json'
    if not file.is_file():
        return None
    with open(file, 'rb') as j:
        rInfo = json.load(j)

    if key:
        return rInfo[key]
    else:
        return rInfo

def getMetaEntry(inputDir: str|pathlib.Path, entry_name: str):
    manifest = getMeta(inputDir, key='manifest')
    # get gaze file
    entry = None
    for e in manifest['entries']:
        if e['type'].lower()==entry_name:
            entry = e
            break
    return entry

def checkRecording(inputDir: str|pathlib.Path, recInfo: Recording):
    actualRecInfo = getRecordingInfo(inputDir)
    if actualRecInfo is None or recInfo.name!=actualRecInfo.name:
        raise ValueError(f"A recording with the name \"{recInfo.name}\" was not found in the folder {inputDir}.")

    # make sure caller did not mess with recInfo
    if recInfo.participant!=actualRecInfo.participant:
        raise ValueError(f"A recording with the participant \"{recInfo.participant}\" was not found in the folder {inputDir}.")
    if recInfo.duration!=actualRecInfo.duration:
        raise ValueError(f"A recording with the duration \"{recInfo.duration}\" was not found in the folder {inputDir}.")
    if recInfo.start_time.value!=actualRecInfo.start_time.value:
        raise ValueError(f"A recording with the start_time \"{recInfo.start_time.display}\" was not found in the folder {inputDir}.")


def copyAdhawkRecording(inputDir: pathlib.Path, outputDir: pathlib.Path, copy_scene_video:bool):
    """
    Copy the relevant files from the specified input dir to the specified output dir
    """
    # figure out what the video file is called
    vid_entry   = getMetaEntry(inputDir, 'video')
    srcFile     = inputDir / vid_entry['file_name']
    if copy_scene_video:
        destFile    = outputDir / 'worldCamera.mp4'
        shutil.copy2(str(srcFile), str(destFile))
    else:
        destFile = None

    return srcFile, destFile

def getCameraHardcoded(outputDir: str|pathlib.Path):
    """
    Get camera calibration
    Hardcoded as per info received from AdHawk
    """
    # turn into camera matrix and distortion coefficients as used by OpenCV
    camera = {}
    camera['cameraMatrix'] = np.array([[8.6175611023603130e+02, 0.                    , 6.4220317156609758e+02],
                                       [0.                    , 8.6411314484767183e+02, 3.4611059418088462e+02],
                                       [0.                    , 0.                    , 1.                    ]])
    camera['distCoeff'] = np.array([6.4704736326069179e-01, 6.9842325204621162e+01, -3.8446374749176787e-03, -6.5685769622407693e-03, 3.3239962207009803e+01, 5.0824354805695138e-01, 6.9018441628550974e+01, 3.1191976852198923e+01])
    camera['resolution'] = np.array([1280, 720])
    camera['position'] = np.array([-0.0685, 0.0152028, 0.00340752])*1000  # our positions are in mm, not m
    camera['rotation'] = cv2.Rodrigues(np.radians(np.array([12.000000000000043, 0.0, 0.0])))[0]

    # store to file
    fs = cv2.FileStorage(str(outputDir / 'calibration.xml'), cv2.FILE_STORAGE_WRITE)
    for key,value in camera.items():
        fs.write(name=key,val=value)
    fs.release()

    return camera['resolution']


def formatGazeData(inputDir: str|pathlib.Path, sceneVideoDimensions: list[int], recInfo: Recording):
    """
    load gazedata json file
    format to get the gaze coordinates w/r/t world camera, and timestamps for every frame of video

    Returns:
        - formatted dataframe with cols for timestamp, frame_idx, and gaze data
        - np array of frame timestamps
    """

    # convert the json file to pandas dataframe
    df = csv2df(inputDir, sceneVideoDimensions)

    # read video file, create array of frame timestamps
    frameTimestamps = video_utils.getFrameTimestampsFromVideo(recInfo.get_scene_video_path())

    # return the gaze data df and frame time stamps array
    return df, frameTimestamps


def csv2df(inputDir: str|pathlib.Path, sceneVideoDimensions: list[int]):
    """
    convert the gaze_data.csv file to a pandas dataframe
    """

    vid_entry = getMetaEntry(inputDir, 'video')
    gaze_entry = getMetaEntry(inputDir, 'gaze')

    file = inputDir / gaze_entry['file_name']
    df = pd.read_csv(file)

    # prepare data frame
    # remove unneeded columns
    df=df.drop(columns=['Screen_X', 'Screen_Y', 'UTC_Time', 'Image_One_Degree_X', 'Image_One_Degree_Y'],errors='ignore') # drop these columns if they exist

    # rename and reorder columns
    lookup = {'Timestamp': 'timestamp',
               'Frame_Index':'frame_idx',
               'Image_X':'gaze_pos_vid_x',
               'Image_Y':'gaze_pos_vid_y',
               'Gaze_X_Left':'gaze_dir_l_x',
               'Gaze_Y_Left':'gaze_dir_l_y',
               'Gaze_Z_Left':'gaze_dir_l_z',
               'Gaze_X_Right':'gaze_dir_r_x',
               'Gaze_Y_Right':'gaze_dir_r_y',
               'Gaze_Z_Right':'gaze_dir_r_z',
               'Gaze_X':'gaze_pos_3d_x',
               'Gaze_Y':'gaze_pos_3d_y',
               'Gaze_Z':'gaze_pos_3d_z',}
    df=df.rename(columns=lookup)
    # reorder
    idx = [lookup[k] for k in lookup if lookup[k] in df.columns]
    idx.extend([x for x in df.columns if x not in idx])   # append columns not in lookup
    df = df[idx]

    # get gaze vector origins
    pp_entry = getMetaEntry(inputDir, 'pupil_position')
    file = inputDir / pp_entry['file_name']
    dfP = pd.read_csv(file)
    dfP = dfP.drop(columns=['UTC_Time'],errors='ignore') # drop these columns if they exist
    # rename and reorder columns
    lookup = {'Timestamp': 'timestamp',
               'Pupil_Pos_X_Left':'gaze_ori_l_x',
               'Pupil_Pos_Y_Left':'gaze_ori_l_y',
               'Pupil_Pos_Z_Left':'gaze_ori_l_z',
               'Pupil_Pos_X_Right':'gaze_ori_r_x',
               'Pupil_Pos_Y_Right':'gaze_ori_r_y',
               'Pupil_Pos_Z_Right':'gaze_ori_r_z',}
    dfP = dfP.rename(columns=lookup)
    # reorder
    idx = [lookup[k] for k in lookup if lookup[k] in dfP.columns]
    idx.extend([x for x in dfP.columns if x not in idx])   # append columns not in lookup
    dfP = dfP[idx]

    # merge
    df = pd.merge(df, dfP, on='timestamp')

    # convert timestamps from s to ms and set as index
    df.loc[:,'timestamp'] *= 1000.0
    # set first gaze timestamp to 0 and express gaze timestamps in video time
    df.loc[:,'timestamp'] -= (df.loc[0,'timestamp'] - (gaze_entry['attribute']['start_time_ms'] - vid_entry['attribute']['start_time_ms']))
    # remove data with negative timestamps
    df = df[df.timestamp >= 0]
    df = df.set_index('timestamp')

    # binocular gaze data
    df.loc[:,'gaze_pos_vid_x'] *= sceneVideoDimensions[0]
    df.loc[:,'gaze_pos_vid_y'] *= sceneVideoDimensions[1]

    # adhawk positive z is backward, ours is forward
    df.loc[:,'gaze_ori_l_z'] = -df.loc[:,'gaze_ori_l_z']
    df.loc[:,'gaze_dir_l_z'] = -df.loc[:,'gaze_dir_l_z']
    df.loc[:,'gaze_ori_r_z'] = -df.loc[:,'gaze_ori_r_z']
    df.loc[:,'gaze_dir_r_z'] = -df.loc[:,'gaze_dir_r_z']
    df.loc[:,'gaze_pos_3d_z'] = -df.loc[:,'gaze_pos_3d_z']

    # adhawk positive y is upward, ours is downward
    df.loc[:,'gaze_ori_l_y'] = -df.loc[:,'gaze_ori_l_y']
    df.loc[:,'gaze_dir_l_y'] = -df.loc[:,'gaze_dir_l_y']
    df.loc[:,'gaze_ori_r_y'] = -df.loc[:,'gaze_ori_r_y']
    df.loc[:,'gaze_dir_r_y'] = -df.loc[:,'gaze_dir_r_y']
    df.loc[:,'gaze_pos_3d_y'] = -df.loc[:,'gaze_pos_3d_y']

    # adhawk gaze pos is in m, ours is in mm
    # NB: gaze ori is in mm!
    df.loc[:,'gaze_pos_3d_x'] *= 1000
    df.loc[:,'gaze_pos_3d_y'] *= 1000
    df.loc[:,'gaze_pos_3d_z'] *= 1000

    # return the dataframe
    return df


import dataclasses
import typing
import pathlib
import json

from .eyetracker import EyeTracker
from .timestamps import Timestamp
from. import utils

@dataclasses.dataclass
class Recording:
    default_json_file_name      : typing.ClassVar[str] = 'recording_info.json'

    name                        : str           = ""
    source_directory            : pathlib.Path  = ""
    working_directory           : pathlib.Path  = ""
    start_time                  : Timestamp     = 0
    duration                    : int           = None
    eye_tracker                 : EyeTracker    = EyeTracker.Unknown
    project                     : str           = ""
    participant                 : str           = ""
    firmware_version            : str           = ""
    glasses_serial              : str           = ""
    recording_unit_serial       : str           = ""
    recording_software_version  : str           = ""
    scene_camera_serial         : str           = ""
    scene_video_file            : str           = ""


    def store_as_json(self, path: str | pathlib.Path):
        path = pathlib.Path(path)
        if path.is_dir():
            path /= self.default_json_file_name
        with open(path, 'w') as f:
            # remove any crap potentially added by subclasses
            to_dump = dataclasses.asdict(self)
            to_dump = {k:to_dump[k] for k in to_dump if k in Recording.__annotations__ and k not in ['working_directory']}      # working_directory will be loaded as the provided path, and shouldn't be stored
            # dump to file
            json.dump(to_dump, f, cls=utils.CustomTypeEncoder, indent=2)

    @staticmethod
    def load_from_json(path: str | pathlib.Path):
        path = pathlib.Path(path)
        if path.is_dir():
            path /= Recording.default_json_file_name
        with open(path, 'r') as f:
            return Recording(**json.load(f, object_hook=utils.json_reconstitute), working_directory=path.parent)


    def get_scene_video_path(self):
        vid = self.working_directory / self.scene_video_file
        if not vid.is_file():
            vid = self.source_directory / self.scene_video_file
        return vid


"""
non_iso.py

This file is meant to contain class definitions for boxes not defined in ISO/IEC 14496-12.
It may be useful to look here:
https://developer.apple.com/library/archive/documentation/QuickTime/QTFF/QTFFChap2/qtff2.html#//apple_ref/doc/uid/TP40000939-CH204-55265

"""
import binascii

from . import iso
from . import mpeglookups
from .util import *
from .core import *


def box_factory_non_iso(fp, header, parent):
    the_box = None
    # Normalise header type so it can be expressed in a Python Class name
    box_type = header.type.replace(" ", "_").replace("-", "_").lower()
    # globals() Return a dictionary representing the current global symbol table
    _box_class = globals().get(box_type.capitalize()+'Box')
    if _box_class:
        the_box = _box_class(fp, header, parent)
    else:
        the_box = UndefinedBox(fp, header, parent)

    return the_box


class UndefinedBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['message'] = 'TODO'
        finally:
            fp.seek(self.start_of_box + self.size)


class Avc1Box(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['num_of_entries'] = read_u32(fp)
            fp.seek(16, 1)
            self.box_info['width'] = read_u16(fp)
            self.box_info['height'] = read_u16(fp)
            self.box_info['horizresolution'] = "{0:#010x}".format(read_u32(fp))
            self.box_info['vertresolution'] = "{0:#010x}".format(read_u32(fp))
            fp.seek(4, 1)
            self.box_info['frame_count'] = read_u16(fp)
            compressorname_size = read_u8(fp)
            self.box_info['compressorname'] = fp.read(compressorname_size).decode('utf-8', errors="ignore")
            padding = fp.read(31 - compressorname_size)
            self.box_info['depth'] = "{0:#06x}".format(read_u16(fp))
            self.box_info['pre_defined'] = read_i16(fp)
            bytes_left = self.start_of_box + self.size - fp.tell()
            while bytes_left > 7:
                current_header = Header(fp)
                current_box = iso.box_factory(fp, current_header, self)
                self.children.append(current_box)
                bytes_left -= current_box.size
        finally:
            fp.seek(self.start_of_box + self.size)


DvheBox = Dvh1Box = DvavBox = Dva1Box = Avc1Box
Hvc1Box = Hev1Box = Av01Box = Avc1Box
Avc4Box = Avc3Box = Avc2Box = Mp4vBox = Avc1Box
EncvBox = Avc1Box


class AvccBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['configuration_version'] = read_u8(fp)
            self.box_info['avc_profile_indication'] = read_u8(fp)
            self.box_info['avc_compatibility'] = read_u8(fp)
            self.box_info['avc_level_indication'] = read_u8(fp)
            self.box_info['lengthSizeMinusOne'] = read_u8(fp) % 4
            self.box_info['numOfSequenceParameterSets'] = read_u8(fp) % 32
            self.box_info['SequenceParameterSets_list'] = []
            for i in range(self.box_info['numOfSequenceParameterSets']):
                spsl = read_u16(fp)
                sequence_param = {
                                    'sequenceParameterSetLength': spsl,
                                    'sequenceParameterSetNALUnit': binascii.b2a_hex(fp.read(spsl)).decode('utf-8')
                                 }
                self.box_info['SequenceParameterSets_list'].append(sequence_param)
            self.box_info['numOfPictureParameterSets'] = read_u8(fp)
            self.box_info['PictureParameterSets_list'] = []
            for i in range(self.box_info['numOfPictureParameterSets']):
                ppsl = read_u16(fp)
                picture_param = {
                                    'pictureParameterSetLength': ppsl,
                                    'pictureParameterSetNALUnit': binascii.b2a_hex(fp.read(ppsl)).decode('utf-8')
                                }
                self.box_info['PictureParameterSets_list'].append(picture_param)
            if (self.box_info['avc_profile_indication'] == 100 or self.box_info['avc_profile_indication'] == 110 or
                self.box_info['avc_profile_indication'] == 122 or self.box_info['avc_profile_indication'] == 144) \
                    and (self.start_of_box + self.size - fp.tell()) > 7:
                self.box_info['chroma_format'] = read_u8(fp) % 4
                self.box_info['bit_depth_luma_minus8'] = read_u8(fp) % 8
                self.box_info['bit_depth_chroma_minus8'] = read_u8(fp) % 8
                self.box_info['numOfSequenceParameterSetExtLength'] = read_u8(fp)
                self.box_info['SequenceParameterSetExt_list'] = []
                for i in range(self.box_info['numOfSequenceParameterSetExtLength']):
                    spse = read_u16(fp)
                    sequence_param = {
                        'sequenceParameterSetExtLength': spse,
                        'sequenceParameterSetExtNALUnit': binascii.b2a_hex(fp.read(spse)).decode('utf-8')
                    }
                    self.box_info['SequenceParameterSetExt_list'].append(sequence_param)
        finally:
            fp.seek(self.start_of_box + self.size)


class HvccBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['configuration_version'] = read_u8(fp)
            profile_info = read_u8(fp)
            self.box_info['general_profile_space'] = profile_info >> 6
            self.box_info['general_tier_flag'] = profile_info >> 5 & 1
            self.box_info['general_profile_idc'] = profile_info & 31
            self.box_info['general_profile_compatibility_flags'] = "{0:#010x}".format(read_u32(fp))
            self.box_info['general_constraint_indicator_flags'] = "0x" + binascii.b2a_hex(fp.read(6)).decode('utf-8')
            self.box_info['general_level_idc'] = read_u8(fp)
            self.box_info['min_spatial_segmentation_idc'] = read_u16(fp) % 4096
            self.box_info['parallelismType'] = read_u8(fp) % 4
            self.box_info['chroma_format_idc'] = read_u8(fp) % 4
            self.box_info['bit_depth_luma_minus8'] = read_u8(fp) % 8
            self.box_info['bit_depth_chroma_minus8'] = read_u8(fp) % 8
            self.box_info['avg_frame_rate'] = read_u16(fp)
            fr = read_u8(fp)
            self.box_info['constant_frame_rate'] = fr >> 6
            self.box_info['num_temporal_layers'] = fr >> 3 & 7
            self.box_info['temporal_id_nested'] = fr >> 2 & 1
            self.box_info['length_size_minus1'] = fr & 3
            self.box_info['num_of_arrays'] = read_u8(fp)
            self.box_info['array_list'] = []
            for i in range(self.box_info['num_of_arrays']):
                nt = read_u8(fp)
                nal_dict = {'array_completeness': nt >> 7, 'NAL_unit_type': nt % 64}
                nal_dict['num_nalus'] = read_u16(fp)
                nal_dict['nalu_list'] = []
                for j in range(nal_dict['num_nalus']):
                    nul = read_u16(fp)
                    nal_dict['nalu_list'].append({
                                                    'nal_unit_length': nul,
                                                    'nal_unit': binascii.b2a_hex(fp.read(nul)).decode('utf-8')
                                                })
                self.box_info['array_list'].append(nal_dict)
        finally:
            fp.seek(self.start_of_box + self.size)


class Av1cBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            first_byte = read_u8(fp)
            self.box_info['marker'] = (first_byte >> 7) & 0x01
            self.version = first_byte & 0x7f
            second_byte = read_u8(fp)
            self.box_info['seq_profile'] = (second_byte >> 5) & 0x07
            self.box_info['seq_level_idx_0'] = second_byte & 0x1f
            third_byte = read_u8(fp)
            self.box_info['seq_tier_0'] = (third_byte >> 7) & 0x01
            self.box_info['high_bitdepth'] = (third_byte >> 6) & 0x01
            self.box_info['twelve_bit'] = (third_byte >> 5) & 0x01
            self.box_info['monochrome'] = (third_byte >> 4) & 0x01
            self.box_info['chroma_subsampling_x'] = (third_byte >> 3) & 0x01
            self.box_info['chroma_subsampling_y'] = (third_byte >> 2) & 0x01
            self.box_info['chroma_sample_position'] = third_byte & 0x03
            fourth_byte = read_u8(fp)
            self.box_info['initial_presentation_delay_present'] = (fourth_byte >> 4) & 0x01
            if self.box_info['initial_presentation_delay_present']:
                self.box_info['initial_presentation_delay_minus_1'] = fourth_byte & 0x0f
            bytes_left = self.size - self.header.header_size - 4
            self.box_info['configOBUs'] = binascii.b2a_hex(fp.read(bytes_left)).decode('utf-8')
        finally:
            fp.seek(self.start_of_box + self.size)


class DvccBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['dv_version_major'] = read_u8(fp)
            self.box_info['dv_version_minor'] = read_u8(fp)
            nextTwoBytes = read_u16(fp)
            self.box_info['dv_profile'] = (nextTwoBytes >> 9) & 0x7f
            self.box_info['dv_level'] = (nextTwoBytes >> 3) & 0x3f
            self.box_info['rpu_present_flag'] = (nextTwoBytes >> 2) & 0x01
            self.box_info['el_present_flag'] = (nextTwoBytes >> 1) & 0x01
            self.box_info['bl_present_flag'] = nextTwoBytes & 0x01
            self.box_info['dv_bl_signal_compatibility_id'] = read_u8(fp) >> 4
        finally:
            fp.seek(self.start_of_box + self.size)


DvvcBox = DvccBox

class BtrtBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['bufferSizeDB'] = read_u32(fp)
            self.box_info['maxBitrate'] = read_u32(fp)
            self.box_info['avgBitrate'] = read_u32(fp)
        finally:
            fp.seek(self.start_of_box + self.size)


class PaspBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['hSpacing'] = read_u32(fp)
            self.box_info['vSpacing'] = read_u32(fp)
        finally:
            fp.seek(self.start_of_box + self.size)


class Mp4aBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            fp.seek(6, 1)
            self.box_info['reference_index'] = "{0:#06x}".format(read_u16(fp))
            self.box_info['audio_encoding_version'] = "{0:#06x}".format(read_u16(fp))
            self.box_info['audio_encoding_revision'] = "{0:#06x}".format(read_u16(fp))
            self.box_info['audio_encoding_vendor'] = "{0:#010x}".format(read_u32(fp))
            self.box_info['audio_channel_count'] = read_u16(fp)
            self.box_info['audio_sample_size'] = read_u16(fp)
            self.box_info['audio_compression_id'] = "{0:#06x}".format(read_u16(fp))
            self.box_info['audio_packet_size'] = "{0:#06x}".format(read_u16(fp))
            self.box_info['audio_sample_rate'] = read_u16_16(fp)
            # need to check this is correct
            bytes_left = self.start_of_box + self.size - fp.tell()
            while bytes_left > 7:
                current_header = Header(fp)
                current_box = iso.box_factory(fp, current_header, self)
                self.children.append(current_box)
                bytes_left -= current_box.size
        finally:
            fp.seek(self.start_of_box + self.size)


Ac_3Box = Ec_3Box = EncaBox = Mp4aBox


class EsdsBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            #self.box_info['elementary_stream_descriptor'] = \
            #    binascii.b2a_hex(fp.read(self.size - (self.header.header_size + 4))).decode('utf-8')
            object_dict = {}
            object_dict['tag_id'] = read_u8(fp)
            padding = 0x0
            while read_u8(fp) == 0x80:
                padding = (padding << 8) + 0x80
            object_dict['preamble'] = hex(padding)
            # read last-read byte again
            fp.seek(-1, 1)
            object_dict['payload_length'] = read_u8(fp)
            object_dict['es_id'] = read_u16(fp)
            object_dict['descriptor_flags'] = read_u8(fp)
            object_dict['descriptor_loop'] = []
            # descriptor loop
            while fp.tell() < (self.start_of_box + self.size):
                this_descriptor_dict = {'tag_id': read_u8(fp)}
                padding = 0x0
                while read_u8(fp) == 0x80:
                    padding = (padding << 8) + 0x80
                this_descriptor_dict['preamble'] = hex(padding)
                # read last byte
                fp.seek(-1, 1)
                this_descriptor_dict['payload_length'] = read_u8(fp)
                payload = fp.read(this_descriptor_dict['payload_length'])
                if this_descriptor_dict['tag_id'] == 4:
                    # it is the elementary stream descriptor
                    mp4ra_type = int.from_bytes(payload[0:1], byteorder='big')
                    this_descriptor_dict['mp4ra_registered_type'] = mpeglookups.mp4ra_table[mp4ra_type] \
                        if mp4ra_type in mpeglookups.mp4ra_table else mp4ra_type
                    type_byte = int.from_bytes(payload[1:2], byteorder='big')
                    es_type = type_byte >> 2
                    this_descriptor_dict['es_type'] = mpeglookups.es_table[es_type] \
                        if es_type in mpeglookups.es_table else es_type
                    this_descriptor_dict['upstream_flag'] = type_byte & 2
                    this_descriptor_dict['specific_info_flag'] = type_byte & 1
                    this_descriptor_dict['buffer_size'] = int.from_bytes(payload[2:5], byteorder='big')
                    this_descriptor_dict['max_bitrate'] = int.from_bytes(payload[5:9], byteorder='big')
                    this_descriptor_dict['avg_bitrate'] = int.from_bytes(payload[9:13], byteorder='big')
                    if this_descriptor_dict['specific_info_flag']:
                        # decoder-specific descriptor embedded in es descriptor
                        decoder_specific_dict = {'tag_id': int.from_bytes(payload[13:14], byteorder='big')}
                        padding = 0x0
                        b = 14
                        while int.from_bytes(payload[b:b+1], byteorder='big') == 0x80:
                            padding = (padding << 8) + 0x80
                            b += 1
                        decoder_specific_dict['preamble'] = hex(padding)
                        decoder_specific_dict['payload_length'] = int.from_bytes(payload[b:b+1], byteorder='big')
                        if es_type == 5:
                            two_bytes = int.from_bytes(payload[b+1:b+3], byteorder='big')
                            field_size = 5
                            sound_codec= two_bytes >> (16 - field_size)
                            bits_read = field_size
                            if sound_codec == 31:
                                field_size = 6
                                sound_codec = ((two_bytes >> (16 - bits_read - field_size)) & 63) + 32
                                bits_read += field_size
                            decoder_specific_dict['sound_codec'] = mpeglookups.sound_codec_table[sound_codec]
                            field_size = 4
                            decoder_specific_dict['freq_id'] = mpeglookups.freq_table[
                                (two_bytes >> (16 - bits_read - field_size)) & 15]
                            bits_read += field_size
                            if decoder_specific_dict['freq_id'] == 15:
                                # we need to read freq directly, the bit-packing makes this a bit cumbersome
                                four_bytes = int.from_bytes(payload[b+2:b+6], byteorder='big')
                                field_size = 24
                                # subtract 8 from bits_read because we are not including first byte in payload
                                bits_read -= 8
                                decoder_specific_dict['freq_value'] = four_bytes >> (32 - bits_read - field_size) & 16777215
                                # adjust to two_bytes
                                bits_read += field_size - 16
                                two_bytes = int.from_bytes(payload[b+4:b+6], byteorder='big')
                            field_size = 4
                            sound_chan = (two_bytes >> (16 - bits_read - field_size)) & 15
                            decoder_specific_dict['channels'] = mpeglookups.sound_channel_table[sound_chan] \
                                if sound_chan in mpeglookups.sound_channel_table else sound_chan
                        else:
                            decoder_specific_dict['payload'] = binascii.b2a_hex(payload[b+1:]).decode('utf-8')
                        this_descriptor_dict['decoder_specific_descriptor'] = decoder_specific_dict
                else:
                    this_descriptor_dict['payload'] = binascii.b2a_hex(payload).decode('utf-8')
                object_dict['descriptor_loop'].append(this_descriptor_dict)
            self.box_info['object_descriptor'] = object_dict
        finally:
            fp.seek(self.start_of_box + self.size)


class Dac3Box(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            my_data = struct.unpack('>I', b'\0' + fp.read(3))[0]
            self.box_info['fscod'] = my_data >> 22
            self.box_info['bsid'] = my_data >> 17 & 31
            self.box_info['bsmod'] = my_data >> 14 & 7
            self.box_info['acmod'] = my_data >> 11 & 7
            self.box_info['lfeon'] = my_data >> 10 & 1
            self.box_info['bit_rate_code'] = my_data >> 5 & 31
        finally:
            fp.seek(self.start_of_box + self.size)


class Dec3Box(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            my_data_sub = read_u16(fp)
            self.box_info['data_rate'] = my_data_sub >> 3
            self.box_info['num_ind_sub'] = my_data_sub & 7
            self.box_info['ind_sub_list'] = []
            for i in range(self.box_info['num_ind_sub']):
                in_s = read_u16(fp)
                fscod = in_s >> 14
                bsid = in_s >> 9 & 31
                asvc = in_s >> 7 & 1
                bsmod = in_s >> 4 & 7
                acmod = in_s >> 1 & 7
                lfeon = in_s & 1
                dep_s = read_u8(fp)
                num_dep_sub = dep_s >> 1 & 15
                bit_9 = dep_s & 1
                sub_dict = {'fscod': fscod, 'bsid': bsid, 'asvc': asvc, 'bsmod': bsmod, 'acmod': acmod,
                            'lfeon': lfeon, 'num_dep_sub': num_dep_sub}
                if num_dep_sub > 0:
                    sub_dict['chan_loc'] = (bit_9 * 512) + read_u8(fp)
                self.box_info['ind_sub_list'].append(sub_dict)
        finally:
            fp.seek(self.start_of_box + self.size)


class DataBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            fp.seek(4, 1)
            self.box_info['text'] = fp.read(self.size - (self.header.header_size + 8)).decode('utf-8')
        finally:
            fp.seek(self.start_of_box + self.size)


class PsshBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['system_id'] = binascii.b2a_hex(fp.read(16)).decode('utf-8')
            if self.box_info['system_id'] == '1077efecc0b24d02ace33c1e52e2fb4b':
                # it is cenc
                self.box_info['key_count'] = read_u32(fp)
                self.box_info['key_list'] = []
                for i in range(self.box_info['key_count']):
                    self.box_info['key_list'].append(binascii.b2a_hex(fp.read(16)).decode('utf-8'))
        finally:
            fp.seek(self.start_of_box + self.size)


class TencBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['reserved'] = read_u8(fp)
            byte_block = read_u8(fp)
            if self.version == 1:
                self.box_info['default_crypt_byte_block'] = byte_block >> 4
                self.box_info['default_skip_byte_block'] = byte_block & 15
            self.box_info['default_isProtected'] = read_u8(fp)
            self.box_info['default_Per_Sample_IV_Size'] = read_u8(fp)
            self.box_info['default_KID'] = binascii.b2a_hex(fp.read(16)).decode('utf-8')
            if self.box_info['default_isProtected'] == 1 and self.box_info['default_Per_Sample_IV_Size'] == 0:
                self.box_info['default_constant_IV_size'] = read_u8(fp)
                self.box_info['default_constant_IV'] = binascii.b2a_hex(fp.read(self.box_info['default_constant_IV_size'])).decode('utf-8')
        finally:
            fp.seek(self.start_of_box + self.size)


class SencBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            # to read this box, we need to know IV size
            self.box_info['sample_count'] = read_u32(fp)
            iv_size_guess = ((self.size - 16) // self.box_info['sample_count'])
            if self.flags & 0x000002 != 0x000002:
                # if no sub-sampling, assume IV has a fixed size 8 or 16 bytes
                iv_size = 16 if (self.box_info['sample_count'] * 16) <= (self.size - 16) else 8
                self.box_info['sample_list'] = []
                for i in range(self.box_info['sample_count']):
                    self.box_info['sample_list'].append({'iv': binascii.b2a_hex(fp.read(iv_size)).decode('utf-8')})
        finally:
            fp.seek(self.start_of_box + self.size)

    def populate_sample_table(self, fp, iv_size):
        # called if sub-sampling used, once we've determined IV size (N.B. assumes IV size constant)
        try:
            fp_orig = fp.tell()
            # move fp to start of list
            fp.seek(self.start_of_box + self.header.header_size + 8)
            self.box_info['sample_list'] = []
            for i in range(self.box_info['sample_count']):
                sample_dict = {'iv': binascii.b2a_hex(fp.read(iv_size)).decode('utf-8')} if iv_size else {}
                sample_dict['subsample_count'] = read_u16(fp)
                sample_dict['subsample_list'] = []
                for j in range(sample_dict['subsample_count']):
                    sample_dict['subsample_list'].append({
                        'BytesOfClearData': read_u16(fp),
                        'BytesOfEncryptedData': read_u32(fp)
                    })
                self.box_info['sample_list'].append(sample_dict)
        finally:
            fp.seek(fp_orig)


class GminBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['graphics_mode'] = read_u16(fp)
            self.box_info['op_color'] = {'red': read_u16(fp), 'green': read_u16(fp), 'blue': read_u16(fp)}
            self.box_info['balance'] = read_u16(fp)
        finally:
            fp.seek(self.start_of_box + self.size)


class KeysBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['entry_count'] = read_u32(fp)
            self.box_info['entry_list'] = []
            for k in range(self.box_info['entry_count']):
                ksize = read_u32(fp)
                self.box_info['entry_list'].append({
                    'key_index': k + 1,
                    'key_size': ksize,
                    'key_namespace': fp.read(4).decode('utf-8'),
                    'key_value': fp.read(ksize -8).decode('utf-8')
                })
        finally:
            fp.seek(self.start_of_box + self.size)


class IlstBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            bytes_left = self.size - self.header.header_size
            while bytes_left > 7:
                current_header = Header(fp)
                if not current_header.type.isprintable():
                    current_header.type = "#{0:#d}".format(struct.unpack('>I', current_header.type.encode('utf-8'))[0])
                # create box directly, not through box factory
                current_box = ItemBox(fp, current_header, self)
                self.children.append(current_box)
                bytes_left -= current_box.size
        finally:
            fp.seek(self.start_of_box + self.size)


class ItemBox(Mp4Box):
    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            bytes_left = self.size - self.header.header_size
            while bytes_left > 7:
                current_header = Header(fp)
                current_box = iso.box_factory(fp, current_header, self)
                self.children.append(current_box)
                bytes_left -= current_box.size
        finally:
            fp.seek(self.start_of_box + self.size)


class IodsBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            object_dict = {}
            object_dict['tag_id'] = read_u8(fp)
            padding = 0x0
            while read_u8(fp) == 0x80:
                padding = (padding << 8) + 0x80
            object_dict['preamble'] = hex(padding)
            # read last-read byte again
            fp.seek(-1, 1)
            object_dict['payload_length'] = read_u8(fp)
            object_dict['od_id'] = read_u16(fp)
            object_dict['od_profile_level'] = read_u8(fp)
            object_dict['scene_profile_level'] = read_u8(fp)
            object_dict['audio_profile_level'] = read_u8(fp)
            object_dict['video_profile_level'] = read_u8(fp)
            object_dict['graphics_profile_level'] = read_u8(fp)
            # unsure if further optional tags can exist
            self.box_info['initial_object_descriptor'] = object_dict
        finally:
            fp.seek(self.start_of_box + self.size)

class Tx3gBox(Mp4FullBox):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['displayFlags'] = read_u32(fp)
            self.box_info['horizontal_justification'] = read_i8(fp)
            self.box_info['vertical_justification'] = read_i8(fp)
            self.box_info['background_color_rgba'] = {'red': read_u8(fp), 'green': read_u8(fp), 'blue': read_u8(fp), 'alpha': read_u8(fp)}
            read_u32(fp)
            self.box_info['default_text_box'] = {'top': read_i16(fp), 'left': read_i16(fp), 'bottom': read_i16(fp), 'right': read_i16(fp)}
            self.box_info['default_style'] = {
                'startChar': read_u16(fp),
                'endChar': read_u16(fp),
                'font_ID': read_u16(fp),
                'face_style_flags': read_u8(fp),
                'font_size': read_u8(fp)
            }
            self.box_info['text_color_rgba'] = {'red': read_u8(fp), 'green': read_u8(fp), 'blue': read_u8(fp), 'alpha': read_u8(fp)}
            current_header = Header(fp)
            current_box = iso.box_factory(fp, current_header, self)
            self.children.append(current_box)
        finally:
            fp.seek(self.start_of_box + self.size)

class FtabBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['entry_count'] = read_u16(fp)
            self.box_info['entry_list'] = []
            for i in range(self.box_info['entry_count']):
                fontID = read_u16(fp)
                fnl = read_u8(fp)
                self.box_info['entry_list'].append({
                    'font_ID': fontID,
                    'font_name_length': fnl,
                    'font_name': fp.read(fnl).decode('utf-8')
                })
        finally:
            fp.seek(self.start_of_box + self.size)

class XyzBox(Mp4Box):

    def __init__(self, fp, header, parent):
        super().__init__(fp, header, parent)
        try:
            self.box_info['data_size'] = read_u16(fp)
            lang = read_u16(fp)
            if lang == 0:
                self.box_info['language'] = '0x00'
            else:
                ch1 = str(chr(96 + (lang >> 10 & 31)))
                ch2 = str(chr(96 + (lang >> 5 & 31)))
                ch3 = str(chr(96 + (lang & 31)))
                self.box_info['language'] = ch1 + ch2 + ch3
            self.box_info['data'] = fp.read(self.box_info['data_size']).decode('utf-8', errors="ignore")
        finally:
            fp.seek(self.start_of_box + self.size)



import setuptools
import runpy

info = runpy.run_path("src/glassesTools/version.py")

with open('README.md') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

required = []

# get required packages
for line in requirements:
    required.append(line)

setuptools.setup(
    name=info['__title__'],
    version=info['__version__'],
    author=info['__author__'],
    author_email=info['__email__'],
    description=info['__description__'],
    long_description=readme,
    long_description_content_type="text/markdown",
    url=info['__url__'],
    project_urls={
        "Source Code": info['__url__'],
    },
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    license=license,
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    include_package_data=True,
    python_requires=">=3.10",
    install_requires=required,
    extras_require = {
        'GUI':  ["imgui_bundle==1.3.0"]
    }
)
